# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import numpy as np
from rdkit import Chem
from rdkit.Chem import PandasTools
from moleculekit.smallmol.smallmol import SmallMol
import gzip
import logging


logger = logging.getLogger(__name__)


def smiReader(file, removeHs, fixHs, isgzip=False):
    from tqdm import tqdm

    if isgzip:
        with gzip.open(file, "rb") as f:
            return smiReader(f, removeHs=removeHs, fixHs=fixHs, isgzip=False)

    if isinstance(file, str):
        with open(file, "r") as f:
            return smiReader(f, removeHs, fixHs)

    decode = False
    if isinstance(file, gzip.GzipFile):
        decode = True

    lines = file.readlines()[1:]
    mols = []
    for i, line in enumerate(tqdm(lines)):
        if decode:
            line = line.decode("utf-8")
        smi, name = line.strip().split()
        try:
            mols.append(SmallMol(smi, removeHs=removeHs, fixHs=fixHs))
        except Exception as e:
            logger.warning(
                "Failed to load molecule with name {} with error {}. Skipping to next molecule.".format(
                    name, e
                )
            )
    return mols


def sdfReader(file, removeHs, fixHs, sanitize, isgzip=False):
    from tqdm import tqdm
    from moleculekit.util import tempname

    if isgzip:
        with gzip.open(file, "rb") as f:
            # SDMolSupplier does not support file handles, need to write temp file
            file = tempname(suffix=".sdf")
            with open(file, "wb") as fout:
                fout.write(f.read())

    supplier = Chem.SDMolSupplier(file, removeHs=removeHs, sanitize=sanitize)
    mols = []
    countfailed = 0
    for mol in tqdm(supplier):
        if mol is None:
            countfailed += 1
            continue
        try:
            mols.append(SmallMol(mol, removeHs=removeHs, fixHs=fixHs))
        except Exception:
            if mol.HasProp("_Name"):
                name = mol.GetProp("_Name")
            countfailed += 1
            logger.warning(
                f"Failed to load molecule with name {name}. Skipping to next molecule."
            )
    if countfailed:
        logger.info(f"Failed to load {countfailed}/{len(supplier)} molecules")
    return mols


class SmallMolLib(object):
    """
    Class to manage ligands databases (sdf). Ligands are stored as moleculekit.smallmol.smallmol.SmallMol objects and
    fields type in the sdf are stored in a list

    Parameters
    ----------
    lib_file: str
        The sdf or smi file path
    removeHs: bool
        If True, the hydrogens of the molecules will be removed
    fixHs: bool
        If True, the hydrogens are added and optimized

    Example
    -------
    >>> import os
    >>> from moleculekit.home import home
    >>> lib = SmallMolLib(os.path.join(home(dataDir='test-smallmol'), 'fda_drugs_light.sdf'))
    >>> lib.numMols
    100

    .. rubric:: Methods
    .. autoautosummary:: moleculekit.smallmol.smallmol.SmallMolLib
       :methods:
    .. rubric:: Attributes
    .. autoautosummary:: moleculekit.smallmol.smallmol.SmallMolLib
       :attributes:

    Attributes
    ----------
    numMols: int
        Number of SmallMol molecules

    """

    def __init__(
        self, libfile=None, removeHs=False, fixHs=True, sanitize=True
    ):  # , n_jobs=1
        if libfile is not None:
            self._mols = self._loadLibrary(
                libfile, removeHs=removeHs, fixHs=fixHs, sanitize=sanitize, ext=None
            )

    def _loadLibrary(
        self, libfile, removeHs=False, fixHs=True, sanitize=True, ext=None
    ):
        isgzip = False
        if ext is None:
            ext = os.path.splitext(libfile)[-1]
        if ext == ".gz":
            isgzip = True
            ext = os.path.splitext(os.path.splitext(libfile)[-2])[-1]

        if ext == ".sdf":
            return sdfReader(libfile, removeHs, fixHs, sanitize, isgzip)
        elif ext == ".smi":
            return smiReader(libfile, removeHs, fixHs, isgzip)
        else:
            raise RuntimeError(f"Invalid file extension {ext}. Could not read it.")

    @property
    def numMols(self):
        """
        Returns the number of molecules
        """
        return len(self._mols)

    def getMols(self, ids=None):
        """
        Returns the SmallMol objects that corresponds ot the indexes of the list passed

        Parameters
        ----------
        ids: list
            The index list of the molecules to return

        Returns
        -------
        smallmollist: list
            The list of SmallMol objects

        Example
        -------
        >>> lib2 = lib.getMols([1,2,3])
        >>> len(lib2)
        3
        """

        if ids is None:
            return self._mols
        if not isinstance(ids, list):
            raise TypeError(f"The argument ids {type(ids)} should be list")

        return np.array(self._mols)[ids]

    def writeSdf(self, sdf_name, fields=None):
        """
        Writes an sdf file with molecules stored. Is it possible also to manage which field will be written

        Parameters
        ----------
        sdf_name: str
            The ouput sdf filename
        fields: list
            A list of the fields to write. If None all are saved
        """

        from rdkit.Chem import SDWriter

        writer = SDWriter(sdf_name)
        if fields is not None:
            if not isinstance(fields, list):
                raise TypeError(f"The fields argument {type(fields)} should be a list")
            writer.SetProps(fields)

        for m in self._mols:
            writer.write(m._mol)

    def writeSmiles(self, smi_name, explicitHs=True, names=False, header=None):
        """
        Writes a smi file with molecules stored. Is it possible to specify the header of the smi file. The name of the
        ligands can be their ligand name or a sequential ID.

        Parameters
        ----------
        smi_name: str
            The ouput smi filename
        names: bool
            Set as True to use the own ligand name for each ligand. Otherwise a sequential ID will be used
        header: str
            The header of the smi file. If is None the smi filename will be used.
        """

        smi_name = os.path.splitext(smi_name)[0] + ".smi"

        f = open(smi_name, "w")

        if header is None:
            header = os.path.splitext(smi_name)[0]

        f.write(header + "\n")

        for n, sm in enumerate(self.getMols()):
            smi = sm.toSMILES(explicitHs=explicitHs)
            name = n if not names else sm.ligname
            f.write(smi + f" {name} \n")

        f.close()

    def appendSmallLib(self, smallLib, strictField=False, strictDirection=1):
        """
        Merge two moleculekit.smallmol.smallmol.SmallMolLib objects

        Parameters
        ----------
        smallLib: moleculekit.smallmol.smallmol.SmallMolLib
            The new SmallMolLib to merge
        """
        self._mols += smallLib._mols

    def appendSmallMol(self, smallmolecule, strictField=False, strictDirection=0):
        """
        Adds a new moleculekit.smallmol.smallmol.SmallMol object in the current SmallMolLib object

        Parameters
        ---------
        smallmol: moleculekit.smallmol.smallmol.SmallMol
            The SmallMol object to add
        """
        self._mols = np.append(self._mols, smallmolecule)

    def removeMols(self, ids):
        """
        Removes the moleculekit.smallmol.smallmol.SmallMol object based on the indexes in the list

        Parameters
        ----------
        ids: list
            The list of molecules index to remove from the SmallMolLib
        """

        if not isinstance(ids, list):
            raise TypeError(
                f"The argument ids {type(ids)} is not valid. Should be list"
            )
        _oldNumMols = self.numMols
        self._mols = np.delete(self._mols, ids)

        logger.warning(
            "[num mols before deleting: {}]. The molecules {} were removed, now the number of "
            "molecules are {} ".format(_oldNumMols, ids, self.numMols)
        )

    def toDataFrame(self, fields=None, molAsImage=True, sketch=True):
        """
        Returns a pandas.DataFrame of the SmallMolLib object.

        Parameters
        ----------
        fields: list
            The list of fields to convert into a pandas DataFrame column
        molAsImage: bool
            If True, the rdkit.Chem.rdchem.Mol is converted into an image
        sketch: bool
            If True, the molecule are rendered to be 2D

        Returns
        -------
        dataframe: pandas.DataFrame
            The pandas DataFrame
        """
        from rdkit.Chem.AllChem import Compute2DCoords
        import pandas as pd
        from copy import deepcopy

        if fields is not None:
            if not isinstance(fields, list):
                raise TypeError(
                    f"The argument fields passed {type(fields)} should be a list "
                )
        else:
            fields = ["ligname", "_mol"] if molAsImage else ["ligname"]

        records = []
        indexes = []
        for i, m in enumerate(self._mols):
            row = dict((f, m.__getattribute__(f)) for f in fields)
            if sketch:
                mm = deepcopy(m._mol)
                Compute2DCoords(mm)
                row["_mol"] = mm
            records.append(row)
            indexes.append(i)

        df = pd.DataFrame(records, columns=fields, index=indexes)
        if molAsImage:
            PandasTools.ChangeMoleculeRendering(df)
        return df

    def copy(self):
        """
        Returns a copy of the SmallMolLib object
        """
        from copy import deepcopy

        return deepcopy(self)

    def __len__(self):
        return len(self._mols)

    def __getitem__(self, item):
        return self._mols[item]

    def __iter__(self):
        _mols = self._mols
        for smallmol in _mols:
            yield smallmol

    def __str__(self):
        _mols = self._mols

        return (
            "Stack of Small molecules."
            "\n\tContains {} Molecules."
            '\n\tSource file: "{}".'
        ).format(len(_mols), self._sdffile)

    def depict(
        self,
        ids=None,
        sketch=True,
        filename=None,
        ipython=False,
        optimize=False,
        optimizemode="std",
        removeHs=True,
        legends=None,
        highlightAtoms=None,
        mols_perrow=3,
    ):

        """
        Depicts the molecules into a grid. It is possible to save it into an svg file and also generates a
        jupiter-notebook rendering

        Parameters
        ----------
        ids: list
            The index of the molecules to depict
        sketch: bool
            Set to True for 2D depiction
        filename: str
            Set the filename for the svg file
        ipython: bool
            Set to True to return the jupiter-notebook rendering
        optimize: bool
            Set to True to optimize the conformation. Works only with 3D.
        optimizemode: ['std', 'mmff']
            Set the optimization mode for 3D conformation
        removeHs: bool
            Set to True to hide hydrogens in the depiction
        legends: str
            The name to used for each molecule. Can be 'names':the name of themselves; or 'items': a incremental id
        highlightAtoms: list
            A List of atom to highligh for each molecule. It can be also a list of atom list, in this case different
            colors will be used
        mols_perrow: int
            The number of molecules to depict per row of the grid

        Returns
        -------
            ipython_svg: SVG object if ipython is set to True

        """
        from rdkit.Chem.AllChem import (
            Compute2DCoords,
            EmbedMolecule,
            MMFFOptimizeMolecule,
            ETKDG,
        )
        from rdkit.Chem import RemoveHs
        from moleculekit.smallmol.util import depictMultipleMols

        if sketch and optimize:
            raise ValueError(
                "Impossible to use optmization in  2D sketch representation"
            )

        if legends is not None and legends not in ["names", "items"]:
            raise ValueError('The "legends" should be "names" or "items"')

        _smallmols = self.getMols(ids)

        if ids is None:
            _mols = [m._mol for m in self._mols]
        else:
            _mols = [m._mol for m in self.getMols(ids)]

        if highlightAtoms is not None:
            if len(highlightAtoms) != len(_mols):
                raise ValueError(
                    "The highlightAtoms {} should have the same length of the "
                    "mols {}".format(len(highlightAtoms), len(_mols))
                )

        if sketch:
            for _m in _mols:
                Compute2DCoords(_m)

        if removeHs:
            _mols = [RemoveHs(_m) for _m in _mols]

        # activate 3D coords optimization
        if optimize:
            if optimizemode == "std":
                for _m in _mols:
                    EmbedMolecule(_m)
            elif optimizemode == "mmff":
                for _m in _mols:
                    MMFFOptimizeMolecule(_m, ETKDG())

        legends_list = []
        if legends == "names":
            legends_list = [_m.getProp("ligname") for _m in _smallmols]
        elif legends == "items":
            legends_list = [str(n + 1) for n in range(len(_smallmols))]

        return depictMultipleMols(
            _mols,
            ipython=ipython,
            legends=legends_list,
            highlightAtoms=highlightAtoms,
            filename=filename,
            mols_perrow=mols_perrow,
        )


SmallMolStack = SmallMolLib

if __name__ == "__main__":
    import doctest
    from moleculekit.home import home

    lib = SmallMolLib(
        os.path.join(home(dataDir="test-smallmol"), "fda_drugs_light.sdf")
    )
    doctest.testmod(extraglobs={"lib": lib})



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from os import path
import os
import shutil
import unittest
from subprocess import call
from moleculekit.util import tempname, natsorted
from moleculekit.molecule import Molecule
from glob import glob
import logging
import platform
import sys

logger = logging.getLogger(__name__)


def dock(
    protein,
    ligand,
    center=None,
    extent=None,
    numposes=20,
    babelexe="obabel",
    vinaexe=None,
):
    """Molecular docking, using Vina

    If centre and extent are not provided, docking will be performed over the whole protein

    Parameters
    ----------
    protein : :class:`Molecule <moleculekit.molecule.Molecule>` object
        Molecule object representing the receptor
    ligand : :class:`Molecule <moleculekit.molecule.Molecule>` object
        Molecule object representing the ligand to dock
    center : list
        3-vec centre of of the search bounding box (optional)
    extent : list
        3-vec linear extent of the search bounding box (optional)
    numposes : int
        Number of poses to return. Vina cannot return more than 20 poses.
    babelexe : str
        Path to babel executable.
    vinaexe : str
        Path to AutoDock Vina executable.

    Returns
    -------
    poses
        Molecule object representing the N<10 best poses
    scores
        3x num_poses matrix containing kcal, rmsd lb, rmsd ub

    Examples
    --------
    >>> poses, scoring = dock(protein, ligand)
    >>> poses, scoring = dock(protein, ligand, center=[ 10., 5., 12. ], extent=[ 15., 15., 15. ] )

    """
    if np.size(protein.coords, 2) != 1 or np.size(ligand.coords, 2) != 1:
        raise RuntimeError("Protein and ligand Molecules should be single frames")

    buffer = 10.0  # Angstrom buffer around protein for whole-protein docking
    c_min = np.min(protein.coords, 0).reshape((1, 3))[0]
    c_max = np.max(protein.coords, 0).reshape((1, 3))[0]

    if center is None:
        center = (buffer + (c_max + c_min)) / 2
    if extent is None:
        extent = (c_max - c_min) + buffer

    # babel -i pdb protein.pdb  -o pdbqt protein.pdbqt -xr
    # babel -i pdb ligand.pdb   -o pdbqt ligand.pdbqt -xhn
    # vina --ligand ligand.pdbqt --receptor protein.pdbqt --center_x 0. --center_y 0. --center_z 0. --size_x 60. --size_y 60. --size_z 60 --exhaustiveness 10
    # babel -m -i pdbqt ligand_out.pdbqt -o pdb out_.pdb -xhn

    protein_pdb = tempname(suffix=".pdb")
    ligand_mol2 = tempname(suffix=".mol2")
    output_pdb = tempname(suffix="_.pdb")
    output_prefix = path.splitext(output_pdb)[0]

    protein_pdbqt = tempname(suffix=".pdbqt")
    ligand_pdbqt = tempname(suffix=".pdbqt")
    output_pdbqt = tempname(suffix=".pdbqt")

    protein.write(protein_pdb)
    lig2 = ligand.copy()
    # babel does not understand mol2 atomtypes and requires elements instead
    lig2.atomtype = lig2.element
    lig2.write(ligand_mol2)

    # Dirty hack to remove the 'END' line from the PDBs since babel hates it
    with open(protein_pdb, "r") as f:
        lines = f.readlines()
    if "END" in lines[:-1]:
        with open(protein_pdb, "w") as f:
            f.writelines(lines[:-1])
    # End of dirty hack

    try:
        if vinaexe is None:
            suffix = "" if platform.system() != "Windows" else ".exe"
            vinaexe = f"{platform.system()}-vina{suffix}"

        vinaexe = shutil.which(vinaexe, mode=os.X_OK)
        if not vinaexe:
            raise RuntimeError(
                "Could not find vina, or no execute permissions are given"
            )
    except Exception:
        raise RuntimeError("Could not find vina, or no execute permissions are given")

    try:
        babelexe = shutil.which(babelexe, mode=os.X_OK)
        if babelexe is None:
            raise RuntimeError(
                "Could not find babel, or no execute permissions are given"
            )
    except Exception:
        raise RuntimeError("Could not find babel, or no execute permissions are given")

    call(
        [babelexe, "-i", "pdb", protein_pdb, "-o", "pdbqt", "-O", protein_pdbqt, "-xr"]
    )

    babelcmd = [
        babelexe,
        "-i",
        "mol2",
        ligand_mol2,
        "-o",
        "pdbqt",
        "-O",
        ligand_pdbqt,
        "-xn",
        "-xh",
    ]

    if np.all(ligand.charge != 0):
        logger.info("Charges detected in ligand and will be used for docking.")
    else:
        logger.info(
            "Charges were not defined for all atoms. Will guess charges using gasteiger method."
        )
        babelcmd += ["--partialcharge", "gasteiger"]
    call(babelcmd)

    if not path.isfile(ligand_pdbqt):
        raise RuntimeError("Ligand could not be converted to PDBQT")
    if not path.isfile(protein_pdbqt):
        raise RuntimeError("Protein could not be converted to PDBQT")

    call(
        [
            vinaexe,
            "--receptor",
            protein_pdbqt,
            "--ligand",
            ligand_pdbqt,
            "--out",
            output_pdbqt,
            "--center_x",
            str(center[0]),
            "--center_y",
            str(center[1]),
            "--center_z",
            str(center[2]),
            "--size_x",
            str(extent[0]),
            "--size_y",
            str(extent[1]),
            "--size_z",
            str(extent[2]),
            "--num_modes",
            str(numposes),
        ]
    )

    call(
        [
            babelexe,
            "-m",
            "-i",
            "pdbqt",
            output_pdbqt,
            "-o",
            "pdb",
            "-O",
            output_pdb,
            "-xhn",
        ]
    )

    outfiles = natsorted(glob(f"{output_prefix}*.pdb"))

    scoring = []
    poses = []
    for i, ligf in enumerate(outfiles):
        scoring.append(_parseScoring(ligf))
        ll = Molecule(ligf)
        ll.viewname = f"Pose {i}"
        poses.append(ll)

    os.remove(protein_pdb)
    os.remove(ligand_mol2)
    os.remove(protein_pdbqt)
    os.remove(ligand_pdbqt)
    os.remove(output_pdbqt)

    return poses, np.array(scoring)


def _parseScoring(outf):
    kcal = rmsdlb = rmsdub = None
    with open(outf, "r") as f:
        for line in f:
            if line.startswith("REMARK VINA RESULT:"):
                pieces = line.split()
                kcal = float(pieces[3])
                rmsdlb = float(pieces[4])
                rmsdub = float(pieces[5])
                break
    if kcal is None or rmsdlb is None or rmsdub is None:
        raise RuntimeError(f"Could not parse vina output correctly {outf}")
    return kcal, rmsdlb, rmsdub


class _TestDocking(unittest.TestCase):
    @unittest.skipUnless(sys.platform.startswith("linux"), "No vina on OSX yet")
    def test_docking(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from os import path

        protein = Molecule(path.join(home(dataDir="test-docking"), "protein.pdb"))
        ligand = Molecule(path.join(home(dataDir="test-docking"), "ligand.pdb"))
        poses, scoring = dock(protein, ligand)

        assert scoring.shape[0] == len(poses)
        assert scoring.shape[0] <= 20 and scoring.shape[0] > 15
        assert scoring.shape[1] == 3


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging

logger = logging.getLogger(__name__)


class MetricTMscore(Projection):
    """Calculates the TMscore of a set of trajectories to a reference structure

    Parameters
    ----------
    refmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The reference Molecule to which we want to calculate the TMscore.
    trajtmstr : str
        Atom selection string for the trajectories from which to calculate the TMscore.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    reftmstr : str, optional
        Atom selection string for the reference structure from which to calculate the TMscore. If None, it defaults to
        `trajrmsdstr`. See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    centerstr : str, optional
        Atom selection string around which to center the wrapping of the trajectories.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    """

    def __init__(self, refmol, trajtmstr, reftmstr=None):
        super().__init__()

        if reftmstr is None:
            reftmstr = trajtmstr
        self._refmol = refmol
        self._reftmsel = self._refmol.atomselect(reftmstr) & (self._refmol.name == "CA")
        self._trajtmsel = trajtmstr

    def _calculateMolProp(self, mol, props="all"):
        res = {}
        res["trajtmsel"] = mol.atomselect(self._trajtmsel) & (mol.name == "CA")
        if np.sum(res["trajtmsel"]) == 0:
            raise RuntimeError("RMSD atom selection resulted in 0 atoms.")
        return res

    def project(self, mol):
        """Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        from moleculekit.util import molTMscore

        mol = mol.copy()
        trajtmsel = self._getMolProp(mol, "trajtmsel")

        tm, _ = molTMscore(mol, self._refmol, trajtmsel, self._reftmsel)
        return tm[:, np.newaxis]

    def getMapping(self, mol):
        """Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        trajtmsel = self._getMolProp(mol, "trajtmsel")
        from pandas import DataFrame

        types = ["tmscore"]
        indexes = [np.where(trajtmsel)[0]]
        description = ["TMscore to reference structure."]
        return DataFrame(
            {"type": types, "atomIndexes": indexes, "description": description}
        )


import unittest


class _TestMetricTMscore(unittest.TestCase):
    def test_tmscore(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        import numpy as np
        from os import path

        mol = Molecule(
            path.join(home(dataDir="test-projections"), "trajectory", "filtered.pdb")
        )
        mol.read(path.join(home(dataDir="test-projections"), "trajectory", "traj.xtc"))
        ref = mol.copy()

        ref.dropFrames(keep=0)
        mol.dropFrames(keep=np.arange(mol.numFrames - 20, mol.numFrames))

        metr = MetricTMscore(ref, "protein and name CA")
        data = metr.project(mol)

        lasttm = np.array(
            [
                0.9633381,
                0.96441294,
                0.96553609,
                0.96088852,
                0.96288511,
                0.95677591,
                0.96544727,
                0.96359811,
                0.95658912,
                0.96893117,
                0.96623924,
                0.96064913,
                0.96207041,
                0.95947848,
                0.96657048,
                0.95993426,
                0.96543296,
                0.96806875,
                0.96437248,
                0.96144066,
            ],
            dtype=np.float32,
        )
        assert np.all(
            np.abs(data.flatten() - lasttm) < 0.001
        ), "TMscore calculation is broken"


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
from copy import deepcopy
import numpy as np
from rdkit import Chem
from rdkit.Chem.rdchem import HybridizationType, BondType
from collections import defaultdict
from moleculekit.smallmol.util import _depictMol, convertToString
from moleculekit.tools.obabel_tools import openbabelConvert
import logging

logger = logging.getLogger(__name__)


_bondtypes_IdxToType = BondType.values
_bondtypes_TypeToString = defaultdict(lambda: "un")
_bondtypes_TypeToString.update(
    {
        BondType.SINGLE: "1",
        BondType.DOUBLE: "2",
        BondType.TRIPLE: "3",
        BondType.AROMATIC: "ar",
    }
)
_bondtypes_FullStringToType = {
    "SINGLE": BondType.SINGLE,
    "DOUBLE": BondType.DOUBLE,
    "TRIPLE": BondType.TRIPLE,
}
_bondtypes_StringToType = {val: key for key, val in _bondtypes_TypeToString.items()}

_hybridizations_IdxToType = HybridizationType.values
_hybridizations_StringToType = {
    "S": HybridizationType.S,
    "SP": HybridizationType.SP,
    "SP2": HybridizationType.SP2,
    "SP3": HybridizationType.SP3,
}


class SmallMol(object):
    """
    Class to manipulate small molecule structures

    Parameters
    ----------
    mol: rdkit.Chem.rdchem.Mol or filename or smile or moleculekit.smallmol.smallmol.SmallMol
        (i) Rdkit molecule or (ii) Location of molecule file (".pdb"/".mol2") or (iii) a smile string or iv) another
        SmallMol object or v) moleculekit.molecule.Molecule object
    ignore_errors: bool
        If True, errors will not be raised.
    force_reading: bool
        If True, and the mol provided is not accepted, the molecule will be initially converted into sdf
    fixHs: bool
        If True, the missing hydrogens are assigned, the others are correctly assinged into the graph of the molecule
    removeHs: bool
        If True, remove the hydrogens

    Examples
    --------
    >>> import os
    >>> from moleculekit.home import home
    >>> from moleculekit.smallmol.smallmol import SmallMol
    >>> SmallMol('CCO')  # doctest: +SKIP
    >>> SmallMol(os.path.join(home(dataDir='test-smallmol'), 'ligand.pdb'), fixHs=False, removeHs=True )  # doctest: +SKIP
    >>> sm = SmallMol(os.path.join(home(dataDir='test-smallmol'), 'benzamidine.mol2'))
    >>> print(sm)                                     # doctest: +ELLIPSIS
    SmallMol with 18 atoms and 1 conformers
    Atom field - bondtype
    Atom field - charge
    ...

    .. rubric:: Methods
    .. autoautosummary:: moleculekit.smallmol.smallmol.SmallMol
       :methods:
    .. rubric:: Attributes
    .. autoautosummary:: moleculekit.smallmol.smallmol.SmallMol
       :attributes:
    """

    _atom_fields = [
        "idx",
        "name",
        "charge",
        "formalcharge",
        "element",
        "chiral",
        "hybridization",
        "neighbors",
        "bondtype",
        "coords",
        "chiraltags",
    ]

    _mol_fields = ["ligname", "_mol"]

    _dtypes = {
        "idx": int,
        "name": object,
        "charge": np.float32,
        "formalcharge": int,
        "element": object,
        "chiral": object,
        "hybridization": int,
        "neighbors": object,
        "bondtype": object,
        "coords": np.float32,
        "chiraltags": object,
        "bonds": np.uint32,
    }

    def __init__(
        self,
        mol,
        ignore_errors=False,
        force_reading=False,
        fixHs=True,
        removeHs=False,
        verbose=True,
    ):
        self._frame = 0

        self._mol = self._initializeMolObj(mol, force_reading, ignore_errors, verbose)
        if removeHs:
            self.removeHs()
        if fixHs:
            self.addHs(addCoords=True)

    def _initializeMolObj(self, mol, force_reading, ignore_errors, verbose=True):
        """
        Read the input and tries to convert it into a rdkit.Chem.rdchem.Mol obj

        Parameters
        ----------
        mol: str or rdkit.Chem.rdchem.Mol or moleculekit.smallmol.smallmol.SmallMol
            i) rdkit.Chem.rdchem.Mol ii) The path to the pdb/mol2 to load iii) The smile string iv) SmallMol object
            v) moleculekit.molecule.nolecule.Molecule
        force_reading: bool
        If the mol provided is not accepted, the molecule will be initially converted into sdf

        Returns
        -------
        _mol: rdkit.Chem.Molecule object
            The rdkit molecule
        smallMol: moleculekit.smallmol.smallmol.SmallMol
            The smallMol object if SmallMol was passed
        """
        from moleculekit.molecule import Molecule

        # If we are converting a Molecule object to a SmallMol object
        remove = False
        natoms = None
        if isinstance(mol, Molecule):
            natoms = mol.numAtoms
            mol = self._fromMolecule(mol)  # Returns a temp filename
            remove = True

        _mol = None
        if isinstance(mol, Chem.Mol):
            _mol = mol
        elif isinstance(mol, str):
            if os.path.isfile(mol):
                name_suffix = os.path.splitext(mol)[-1]
                # load mol2 file
                if name_suffix == ".mol2":
                    _mol = Chem.MolFromMol2File(mol, removeHs=False)
                # load pdb file
                elif name_suffix == ".pdb":
                    _mol = Chem.MolFromPDBFile(mol, removeHs=False)
                # load single-molecule sdf file
                elif name_suffix == ".sdf":
                    sms = Chem.SDMolSupplier(mol, removeHs=False)
                    if len(sms) != 1:
                        logger.warning(
                            f"More than one molecules found in {mol}. SmallMol will only read the first. If you want to read them all use the SmallMolLib class."
                        )
                    _mol = sms[0]
                # if the file failed to be loaded and 'force_reading' = True, file convert to sdf and than loaded
                if _mol is None and force_reading:
                    logger.warning(f"Reading {mol} with force_reading procedure")
                    sdf = openbabelConvert(mol, name_suffix, "sdf")
                    _mol = Chem.SDMolSupplier(sdf, removeHs=False)[0]
                    os.remove(sdf)
            else:
                # assuming it is a smile
                psmile = Chem.SmilesParserParams()
                psmile.removeHs = False
                _mol = Chem.MolFromSmiles(mol, psmile)

        if remove:  # Remove temp file
            os.remove(mol)

        if _mol is None and not ignore_errors:
            if isinstance(mol, str):
                frerr = (
                    " Try by setting the force_reading option as True."
                    if not force_reading
                    else ""
                )
                raise ValueError(f"Failed to read file {mol}.{frerr}")
            elif isinstance(mol, Molecule):
                raise ValueError("Failed converting Molecule to SmallMol")
            else:
                raise RuntimeError(f"Failed reading molecule {mol}.")

        if natoms is not None and natoms != _mol.GetNumAtoms():
            raise RuntimeError(
                "Number of atoms changed while converting to rdkit molecule"
            )

        return _mol

    @property
    def _idx(self):
        return np.array(
            [a.GetIdx() for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes["idx"]
        )

    @property
    def _element(self):
        return np.array(
            [a.GetSymbol() for a in self._mol.GetAtoms()],
            dtype=SmallMol._dtypes["element"],
        )

    @property
    def _name(self):
        return np.array(
            [f"{a.GetSymbol()}{a.GetIdx()}" for a in self._mol.GetAtoms()],
            dtype=SmallMol._dtypes["name"],
        )

    @property
    def _charge(self):
        return np.array(
            [
                a.GetPropsAsDict()["_TriposPartialCharge"]
                if a.HasProp("_TriposPartialCharge")
                else 0.0
                for a in self._mol.GetAtoms()
            ],
            dtype=SmallMol._dtypes["charge"],
        )

    @property
    def _formalcharge(self):
        return np.array(
            [a.GetFormalCharge() for a in self._mol.GetAtoms()],
            dtype=SmallMol._dtypes["formalcharge"],
        )

    @property
    def _chiral(self):
        return np.array(
            [
                a.GetPropsAsDict()["_CIPCode"] if a.HasProp("_CIPCode") else ""
                for a in self._mol.GetAtoms()
            ],
            dtype=SmallMol._dtypes["chiral"],
        )

    @property
    def _hybridization(self):
        return np.array(
            [int(a.GetHybridization()) for a in self._mol.GetAtoms()],
            dtype=SmallMol._dtypes["hybridization"],
        )

    @property
    def _neighbors(self):
        return np.array(
            [[na.GetIdx() for na in a.GetNeighbors()] for a in self._mol.GetAtoms()],
            dtype=SmallMol._dtypes["neighbors"],
        )

    @property
    def _neighborsbondtype(self):
        return np.array(
            [
                [
                    self._mol.GetBondBetweenAtoms(int(a), int(n)).GetBondType()
                    for n in neighs
                ]
                for a, neighs in enumerate(self._neighbors)
            ],
            dtype=SmallMol._dtypes["bondtype"],
        )

    @property
    def _bonds(self):
        bonds = [
            [bo.GetBeginAtomIdx(), bo.GetEndAtomIdx()] for bo in self._mol.GetBonds()
        ]
        if len(bonds):
            return np.vstack(bonds).astype(SmallMol._dtypes["bonds"])
        else:
            return np.zeros((0, 2), dtype=SmallMol._dtypes["bonds"])

    @property
    def _bondtype(self):
        return np.array(
            [
                _bondtypes_TypeToString[
                    self._mol.GetBondBetweenAtoms(int(b1), int(b2)).GetBondType()
                ]
                for b1, b2 in self._bonds
            ],
            dtype=SmallMol._dtypes["bondtype"],
        )

    @property
    def _chiraltags(self):
        return np.array(
            [a.GetChiralTag() for a in self._mol.GetAtoms()],
            dtype=SmallMol._dtypes["chiraltags"],
        )

    @property
    def ligname(self):
        return self._mol.GetProp("_Name") if self._mol.HasProp("_Name") else "UNK"

    @property
    def _resname(self):
        return np.array(
            [self.ligname.strip() for _ in range(self.numAtoms)], dtype=object
        )

    @property
    def _coords(self):
        if self._mol.GetNumConformers() != 0:
            return np.array(
                np.stack(
                    [cc.GetPositions() for cc in self._mol.GetConformers()], axis=2
                ),
                dtype=SmallMol._dtypes["coords"],
            )
        else:
            return np.zeros((self.numAtoms, 3, 0), dtype=SmallMol._dtypes["coords"])

    @property
    def _atomtype(self):
        return np.array(
            [
                a.GetPropsAsDict()["_TriposAtomType"]
                if a.HasProp("_TriposAtomType")
                else ""
                for a in self._mol.GetAtoms()
            ],
            dtype=object,
        )

    @property
    def frame(self):
        if self._frame < 0 or self._frame >= self.numFrames:
            raise RuntimeError("frame out of range")
        return self._frame

    @frame.setter
    def frame(self, value):
        if value < 0 or ((self.numFrames != 0) and (value >= self.numFrames)):
            raise RuntimeError(
                f"Frame index out of range. Molecule contains {self.numFrames} frame(s). Frames are 0-indexed."
            )
        self._frame = value

    # Dummy properties for compatibility with Molecule
    @property
    def _serial(self):
        return np.arange(self.numAtoms, dtype=np.int32)

    @property
    def _resid(self):
        return np.ones(self.numAtoms, dtype=np.int32)

    @property
    def _occupancy(self):
        return np.zeros(self.numAtoms, dtype=np.float32)

    @property
    def _beta(self):
        return np.zeros(self.numAtoms, dtype=np.float32)

    @property
    def _masses(self):
        return np.zeros(self.numAtoms, dtype=np.float32)

    @property
    def _insertion(self):
        return np.array(["" for _ in range(self.numAtoms)], dtype=object)

    @property
    def _chain(self):
        return np.array(["" for _ in range(self.numAtoms)], dtype=object)

    @property
    def _segid(self):
        return np.array(["" for _ in range(self.numAtoms)], dtype=object)

    @property
    def _altloc(self):
        return np.array(["" for _ in range(self.numAtoms)], dtype=object)

    @property
    def _record(self):
        return np.array(["HETATM" for _ in range(self.numAtoms)], dtype=object)

    @property
    def numAtoms(self):
        return self._mol.GetNumAtoms()

    @property
    def numFrames(self):
        return self._mol.GetNumConformers()

    @property
    def _totalcharge(self):
        return sum(self._formalcharge)

    def _getBonds(self, fileBonds=True, guessBonds=True):
        return self._bonds

    def filter(self, sel):
        # Not implemented
        raise NotImplementedError("Filtering atoms not supported yet.")

    def copy(self):
        """
        Create a copy of the molecule object

        Returns
        -------
        newsmallmol : :class:`SmallMol`
            A copy of the object
        """
        return SmallMol(
            Chem.Mol(self._mol), fixHs=False, removeHs=False
        )  # Chem.Mol creates a deep copy of the C++ object

    def get(self, returnField, sel="all", convertType=True, invert=False):
        """
        Returns the property for the atom specified with the selection. The selection is another atom property

        Parameters
        ----------
        returnField: str
            The field of the atom to return
        sel: str
            The selection string. atom field name followed by spaced values for that field
        convertType: bool
            If True, and where possible the returnField is converted in rdkit object
            Default: True
        invert: bool
            If True, the selection is inverted
            Default: False

        Returns
        -------
        values: np.array
            The array of values for the property

        Example
        -------
        >>> sm.get('element', 'idx 0 1 7')  # doctest: +SKIP
        array(['C', 'C', 'H'],
              dtype='<U1')
        >>> sm.get('hybridization', 'element N')  # doctest: +SKIP
        array([rdkit.Chem.rdchem.HybridizationType.SP2,
               rdkit.Chem.rdchem.HybridizationType.SP2], dtype=object)
        >>> sm.get('hybridization', 'element N', convertType=False)
        array([3, 3])
        >>> sm.get('element', 'hybridization sp2')  # doctest: +SKIP
        array(['C', 'C', 'C', 'C', 'C', 'C', 'C', 'N', 'N'],
              dtype='<U1')
        >>> sm.get('element', 'hybridization S')  # doctest: +SKIP
        array(['H', 'H', 'H', 'H', 'H', 'H', 'H', 'H', 'H'],
              dtype='<U1')
        >>> sm.get('element', 'hybridization 1')  # doctest: +SKIP
        array(['H', 'H', 'H', 'H', 'H', 'H', 'H', 'H', 'H'],
              dtype='<U1')
        >>> sm.get('atomobject', 'element N')  # doctest: +SKIP
        array([<rdkit.Chem.rdchem.Atom object at 0x7faf616dd120>,
               <rdkit.Chem.rdchem.Atom object at 0x7faf616dd170>], dtype=object)
        """
        if sel == "all":
            sel = f"idx {convertToString(self._idx.tolist())}"
        # get the field key and the value to grep
        key = sel.split()[0]
        selector = sel.split()[1:]

        if key not in self._atom_fields:
            raise KeyError(f"The property passed {key} does not exist")
        if len(selector) == 0:
            raise ValueError("No selection was provided")

        # get the returnField and process exceptional field
        if hasattr(self, "_" + key):
            _arrayFrom = self.__getattribute__("_" + key)
        else:
            _arrayFrom = self.__getattribute__(key)

        # special selector for hybridization: can be idx, or rdkit.Chem.rdchem.HybridizationType
        if key == "hybridization":
            try:
                selector = [_hybridizations_StringToType[s.upper()] for s in selector]
            except Exception:
                pass

        _dtype = self._dtypes[key]
        if _dtype is not object:
            selector = [_dtype(s) for s in selector]
        idxs = np.concatenate([np.where(_arrayFrom == s)[0] for s in selector])
        if invert:
            idxs = np.array([i for i in self._idx if i not in idxs])
        idxs = np.sort(idxs)

        if returnField == "atomobject":
            return self.getAtoms()[idxs]
        elif returnField == "bondtype":
            if convertType:
                return np.array(
                    [
                        [_bondtypes_IdxToType[bt] for bt in neighbt]
                        for neighbt in self._neighborsbondtype[idxs]
                    ],
                    dtype=object,
                )
            else:
                return self._neighborsbondtype[idxs]
        elif returnField == "hybridization" and convertType:
            _arrayTo = np.array(
                [_hybridizations_IdxToType[v] for v in self._hybridization],
                dtype=object,
            )
        else:
            _arrayTo = self.__getattribute__("_" + returnField)

        return _arrayTo[idxs]

    def foundBondBetween(self, sel1, sel2, bondtype=None):
        """
        Returns True if at least a bond is found between the two selections. It is possible to check for specific bond
        type. A tuple is returned in the form (bool, [ [(idx1,idx2), rdkit.Chem.rdchem.BondType]] ])

        Parameters
        ----------
        sel1: str
            The selection for the first set of atoms
        sel2: str
            The selection for the second set of atoms
        bondtype: str or int
            The bondtype as index or string
            Default: None

        Returns
        -------
        isbond: bool
            True if a bond was found
        details: list
            A list of lists with the index of atoms in the bond and its type
        """

        if isinstance(bondtype, str):
            _btype = _bondtypes_FullStringToType[bondtype]
        else:
            _btype = bondtype

        atomIdx_sel1 = self.get("idx", sel1)
        neighbors_sel1 = self.get("neighbors", sel1)
        bondtypes_sel1 = self.get("bondtype", sel1)
        atomIdx_sel2 = self.get("idx", sel2)

        founds = []
        for aIdx, neighbors_set, btype_set in zip(
            atomIdx_sel1, neighbors_sel1, bondtypes_sel1
        ):
            for neighbor, btype in zip(neighbors_set, btype_set):
                if neighbor in atomIdx_sel2:
                    btype_str = _bondtypes_TypeToString[btype]
                    if _btype is not None and _btype == btype:
                        founds.append([(aIdx, neighbor), btype_str])
                    elif bondtype is None:
                        founds.append([(aIdx, neighbor), btype_str])
        if len(founds) != 0:
            return True, founds

        return False

    def isChiral(self, returnDetails=False):
        """
        Returns True if the molecule has at least one chiral atom. If returnDetails is set as True,
        a list of tuples with the atom idx and chiral type is returned.

        Parameters
        ----------
        returnDetails: bool
            If True, returns the chiral atoms and their chiral types
            Default: False

        Returns
        -------
        ischiral: bool
            True if the atom has at least a chiral atom
        details: list
            A list of tuple with the chiral atoms and their types

        Example
        -------
        >>> chiralmol.isChiral()  # doctest: +SKIP
        True
        >>> chiralmol.isChiral(returnDetails=True)  # doctest: +SKIP
        (True, [('C2', 'R')])
        """

        _chirals = self._chiral

        idxs = np.where(_chirals != "")[0]
        if len(idxs) == 0:
            return False
        if returnDetails:
            idxs = idxs.astype(str)
            idxs_str = " ".join(idxs)
            names = self.get("name", f"idx {idxs_str}")
            chirals = self.get("chiral", f"idx {idxs_str}")

            return True, [(a, c) for a, c in zip(names, chirals)]

        return True

    def getAtoms(self):
        """
        Retuns an array with the rdkit.Chem.rdchem.Atom present in the molecule
        """
        return np.array([a for a in self._mol.GetAtoms()])

    def getCenter(self):
        """
        Returns geometrical center of molecule conformation
        """
        coords = self._coords[:, :, self.frame]
        return coords.mean(axis=0).astype(np.float32)

    def generateConformers(
        self, num_confs=400, optimizemode="mmff", align=True, append=True
    ):
        """
        Generates ligand conformers

        Parameters
        ----------
        num_confs: int
           Number of conformers to generate.
        optimizemode: str
            The optimizemode to use. Can be  'uff', 'mmff'
            Default: 'mmff'
        align: bool
            If True, the conformer are aligned to the first one
            Default: True
        append: bool
            If False, the current conformers are deleted
            Default: True

        """
        from rdkit.Chem.AllChem import (
            UFFOptimizeMolecule,
            MMFFOptimizeMolecule,
            EmbedMultipleConfs,
        )
        from rdkit.Chem.rdMolAlign import AlignMolConformers

        if not append:
            self.dropFrames(np.arange(self.numFrames))

        # get the rdkit mol and copy it.
        mol = deepcopy(self._mol)
        # hydrogens are added for safety
        mol = Chem.AddHs(mol)

        # generating conformations
        ids = EmbedMultipleConfs(
            mol,
            clearConfs=False,
            numConfs=num_confs,
            pruneRmsThresh=1.0,
            maxAttempts=10000,
        )
        if optimizemode not in ["uff", "mmff"]:
            raise ValueError('Unknown optimizemode. Should be  "uff", "mmff"')
        # optimizing conformations depends on the optimizemode passed
        for id in ids:
            if optimizemode == "mmff":
                MMFFOptimizeMolecule(mol, confId=id)
            elif optimizemode == "uff":
                UFFOptimizeMolecule(mol, confId=id)

        if align:
            AlignMolConformers(mol)

        self._mol = mol

    def align(self, refmol):
        from rdkit.Chem.rdMolAlign import GetO3A
        from moleculekit.molecule import Molecule

        if isinstance(refmol, SmallMol):
            refmol = refmol._mol
        if isinstance(refmol, Molecule):
            refmol = SmallMol(refmol)._mol

        pyO3A = GetO3A(self._mol, refmol)
        rmsd = pyO3A.Align()
        logger.info(f"Alignment with a RMSD of {rmsd}")

    def dropFrames(self, frames="all"):
        if isinstance(frames, int):
            frames = np.array([frames])
        if isinstance(frames, list):
            frames = np.array(frames)
        if isinstance(frames, str) and frames == "all":
            ids = [cc.GetId() for cc in self._mol.GetConformers()]
            for f in ids:
                self._mol.RemoveConformer(int(f))
        else:
            if np.any(frames >= self.numFrames):
                raise RuntimeError(
                    f"Cannot remove more frames than existing which is {self.numFrames}"
                )
            ids = [cc.GetId() for cc in self._mol.GetConformers()]
            for f in frames:
                self._mol.RemoveConformer(int(ids[f]))

    def write(self, fname, frames=None, merge=True):
        ext = os.path.splitext(fname)[-1]
        if ext == ".sdf":
            chemwrite = Chem.SDWriter
            # If merge is set as True a unique file is generated
            if merge:
                writer = chemwrite(fname)
                writer.write(self._mol)
            else:  # If merge is set as False a file is created for each conformer
                if frames is None:
                    frames = list(range(self.numFrames))
                os.makedirs(os.path.dirname(fname), exist_ok=True)
                for i in frames:
                    fname_first = os.path.splitext(fname)[0]
                    currfname = os.path.join(f"{fname_first}_{i}{ext}")
                    writer = chemwrite(currfname)
                    writer.write(self._mol, confId=i)
        else:
            mol = self.toMolecule()
            mol.write(fname, frames)

    def view(self, *args, **kwargs):
        self.toMolecule().view(*args, **kwargs)

    def toSMARTS(self, explicitHs=False):
        """
        Returns the smarts string of the molecule

        Parameters
        ----------
        explicitHs: bool
            Set as True for keep the hydrogens

        Returns
        -------
        smart: str
            The smarts string
        """
        from copy import deepcopy

        rmol = deepcopy(self._mol)
        if not explicitHs and len(np.where(self._element == "H")[0]) != 0:
            rmol = Chem.RemoveHs(rmol)

        return Chem.MolToSmarts(rmol, isomericSmiles=True)

    def toSMILES(self, explicitHs=False, kekulizeSmile=True):
        """
        Returns the smiles string of the molecule

        Parameters
        ----------
        explicitHs: bool
            Set as True to keep the hydrogens
        kekulizeSmile: bool
            Set as True to return the kekule smile format

        Returns
        -------
        smi: str
            The smiles string
        """
        from copy import deepcopy

        rmol = deepcopy(self._mol)

        if not explicitHs and len(np.where(self._element == "H")[0]) != 0:
            rmol = Chem.RemoveHs(rmol)

        if kekulizeSmile:
            Chem.Kekulize(rmol)
            smi = Chem.MolToSmiles(rmol, isomericSmiles=True, kekuleSmiles=True)
        else:
            smi = Chem.MolToSmiles(rmol, isomericSmiles=True)

        return smi

    def _fromMolecule(self, mol):
        """
        Returns the rdkit.Chem.rdchem.Mol object from an moleculekit.molecule.Molecule one

        Parameters
        ----------
        mol: moleculekit.molecule.Molecule
            The moleculekit Molecule object

        Returns
        -------
        rmol: rdkit.Chem.rdchem.Mol
            The rdkit molecule object
        """

        from tempfile import NamedTemporaryFile

        if np.all(mol.bondtype == "un") and np.all(mol.atomtype == ""):
            tmppdb = NamedTemporaryFile(suffix=".pdb").name
            mol.write(tmppdb)
            return tmppdb
        else:
            tmpmol2 = NamedTemporaryFile(suffix=".mol2").name
            mol.write(tmpmol2)
            return tmpmol2

    def toMolecule(self, ids=None):
        """
        Return the moleculekit.molecule.Molecule

        Parameters
        ----------
        ids: list
            The list of conformer ids to store in the moleculekit Molecule object- If None, all are returned
            Default: None

        Returns
        -------
        mol: moleculekit.molecule.Molecule
            The moleculekit Molecule object

        """
        from moleculekit.molecule import Molecule

        class NoConformerError(Exception):
            pass

        if ids is None:
            ids = np.arange(self.numFrames)

        if self.numFrames == 0:
            raise NoConformerError(
                "No Conformers are found in the molecule. Generate at least one conformer."
            )
        elif not isinstance(ids, list) and not isinstance(ids, np.ndarray):
            raise ValueError("The argument ids should be a list of confomer ids")

        mol = Molecule()
        mol.empty(self.numAtoms)
        mol.record[:] = "HETATM"
        mol.resname[:] = self.ligname[:3]
        mol.resid[:] = self._resid
        mol.coords = self._coords[:, :, ids]
        mol.name[:] = self._name
        mol.element[:] = self._element
        mol.formalcharge[:] = self._formalcharge
        mol.charge[:] = self._charge
        mol.box = np.zeros((3, self.numFrames), dtype=np.float32)
        mol.viewname = self.ligname
        mol.bonds = self._bonds
        mol.bondtype = self._bondtype
        mol.atomtype = self._atomtype
        return mol

    def depict(
        self,
        sketch=True,
        filename=None,
        ipython=False,
        optimize=False,
        optimizemode="std",
        removeHs=True,
        atomlabels=None,
        highlightAtoms=None,
        resolution=(400, 200),
    ):
        """
        Depicts the molecules. It is possible to save it into an svg file and also generates a jupiter-notebook rendering

        Parameters
        ----------
        sketch: bool
            Set to True for 2D depiction
        filename: str
            Set the filename for the svg file
        ipython: bool
            Set to True to return the jupiter-notebook rendering
        optimize: bool
            Set to True to optimize the conformation. Works only with 3D.
        optimizemode: ['std', 'mmff']
            Set the optimization mode for 3D conformation
        removeHs: bool
            Set to True to hide hydrogens in the depiction
        atomlabels: str
            Accept any combinations of the following pararemters as unique string '%a%i%c%*' a:atom name, i:atom index,
            c:atom formal charge (+/-), *:chiral (* if atom is chiral)
        highlightAtoms: list
            List of atom to highlight. It can be also a list of atom list, in this case different colors will be used
        resolution: tuple of integers
            Resolution in pixels: (X, Y)

        Returns
        -------
            ipython_svg: SVG object if ipython is set to True

        Example
        -------
        >>> sm.depict(ipython=True, optimize=True, optimizemode='std')  # doctest: +SKIP
        >>> sm.depict(ipython=True, sketch=True)  # doctest: +SKIP
        >>> sm.depict(ipython=True, sketch=True)  # doctest: +SKIP
        >>> sm.depict(ipython=True, sketch=True, atomlabels="%a%i%c")  # doctest: +SKIP
        >>> ids = np.intersect1d(sm.get('idx', 'hybridization SP2'), sm.get('idx', 'element C'))  # doctest: +SKIP
        >>> sm.depict(ipython=True, sketch=True,highlightAtoms=ids.tolist(), removeHs=False)  # doctest: +SKIP
        """
        from rdkit import Chem
        from rdkit.Chem.AllChem import (
            Compute2DCoords,
            EmbedMolecule,
            MMFFOptimizeMolecule,
            ETKDG,
        )
        from copy import deepcopy

        if sketch and optimize:
            raise ValueError(
                "Impossible to use optimization in  2D sketch representation"
            )

        if optimizemode not in ["std", "mmff"]:
            raise ValueError(
                f'Optimization mode {optimizemode} not understood. Can be "std" or "ff"'
            )

        elements = self._element
        indexes = self._idx
        formalcharges = self._formalcharge
        chirals = self._chiral

        _mol = deepcopy(self._mol)

        if sketch:
            Compute2DCoords(_mol)

        if removeHs:
            _mol = Chem.RemoveHs(_mol)
            elements = self.get("element", "element H", invert=True)
            indexes = self.get("idx", "element H", invert=True)
            formalcharges = self.get("formalcharge", "element H", invert=True)
            chirals = self.get("chiral", "element H", invert=True)

        _labelsFunc = ["a", "i", "c", "*"]

        if atomlabels is not None:
            labels = atomlabels.split("%")[1:]
            formalcharges = [
                "" if c == 0 else "+" if c == 1 else "-" for c in formalcharges
            ]
            chirals = ["" if c == "" else "*" for c in chirals]
            values = [elements, indexes, formalcharges, chirals]

            idxs = [_labelsFunc.index(lab) for lab in labels]
            labels_required = [values[i] for i in idxs]
            atomlabels = [
                "".join([str(i) for i in a]) for a in list(zip(*labels_required))
            ]

        if optimize:
            if optimizemode == "std":
                EmbedMolecule(_mol, ETKDG())
            elif optimizemode == "mmff":
                MMFFOptimizeMolecule(_mol)

        return _depictMol(
            _mol,
            filename=filename,
            ipython=ipython,
            atomlabels=atomlabels,
            highlightAtoms=highlightAtoms,
            resolution=resolution,
        )

    def addHs(self, addCoords=True):
        self._mol = Chem.AddHs(self._mol, addCoords=addCoords)

    def removeHs(self):
        self._mol = Chem.RemoveHs(self._mol)

    def __repr__(self):
        return (
            "<{}.{} object at {}>\n".format(
                self.__class__.__module__, self.__class__.__name__, hex(id(self))
            )
            + self.__str__()
        )

    def __str__(self):
        def formatstr(name, field):
            if isinstance(field, np.ndarray) or isinstance(field, list):
                rep = f"{name} shape: {np.shape(field)}"
            elif field == "reps":
                rep = f"{name}: {len(self.reps.replist)}"
            else:
                rep = f"{name}: {field}"
            return rep

        rep = f"SmallMol with {self.numAtoms} atoms and {self.numFrames} conformers"
        for p in sorted(self._atom_fields):
            if p.startswith("_"):
                continue
            rep += "\n"
            rep += f"Atom field - {p}"
        for j in sorted(self.__dict__.keys() - list(SmallMol._atom_fields)):
            if j[0] == "_":
                continue
            rep += "\n"
            rep += formatstr(j, self.__dict__[j])

        return rep


if __name__ == "__main__":
    import doctest
    from moleculekit.home import home
    from moleculekit.smallmol.smallmollib import SmallMolLib

    lib = SmallMolLib(
        os.path.join(home(dataDir="test-smallmol"), "fda_drugs_light.sdf")
    )
    sm = SmallMol(os.path.join(home(dataDir="test-smallmol"), "benzamidine.mol2"))
    doctest.testmod(extraglobs={"lib": lib.copy(), "sm": sm.copy()})



from moleculekit.version import version as _version
from moleculekit.home import home as __home
import os
import logging.config

__version__ = _version()

try:
    logging.config.fileConfig(
        os.path.join(__home(), "logging.ini"), disable_existing_loggers=False
    )
except Exception:
    print("MoleculeKit: Logging setup failed")



import numpy as np


def _wrapBondedDistance(pos, box):
    # Assuming bonds can't cross multiple periodic boxes this is marginally faster
    hbox = box / 2
    under = pos < -hbox
    over = pos > hbox
    pos[under] += box[under]
    pos[over] -= box[over]
    return pos


def dihedralAngle(pos, box=None):
    """Calculates a dihedral angle.

    Parameters
    ----------
    pos: np.ndarray
        An array of 4x3 size where each row are the coordinates of an atom defining the dihedral angle
    box: np.ndarray
        The size of the periodic box

    Returns
    -------
    angle: float
        The angle in radians
    """
    if pos.shape[0] != 4 or pos.shape[1] != 3:
        raise RuntimeError(
            "dihedralAngles requires a 4x3 sized coordinate matrix as input."
        )

    r12 = pos[0] - pos[1]
    r23 = pos[1] - pos[2]
    r34 = pos[2] - pos[3]
    if box is not None and not np.all(box == 0):
        r12 = _wrapBondedDistance(r12, box)
        r23 = _wrapBondedDistance(r23, box)
        r34 = _wrapBondedDistance(r34, box)

    c1 = np.cross(r23, r34, axisa=0, axisb=0, axisc=0)
    c2 = np.cross(r12, r23, axisa=0, axisb=0, axisc=0)

    p1 = (r12 * c1).sum(axis=0)
    p1 *= (r23 * r23).sum(axis=0) ** 0.5
    p2 = (c1 * c2).sum(axis=0)

    return -np.arctan2(p1, p2)
